%
% (c) Microsoft Corporation, 2000
%

Support for compiling Pan effects into DXTransform effects - called
by the backend 'driver' in Compiler.genUI

The 'architecture' of the DXTransform backend is as follows: Using
the ATL-based base class 'CDXBaseNTo1' (provided by the DXMedia SDK),
we generate and insert the per-transform code and data. Apart from
a method that generates the transform (and the WorkProc that calls it),
we also emit an .rgs file which the ATL COM object in-proc wrappers
depend on for registration. 

A DXTransform may support a number of properties through which the
behaviour of the transform can be tweaked. Support these here by
generating the property set/get'ers, along with the IDL that
describes the component we're generating.

In time, we'll also support the generation of property pages which
allows author/design-time customisation of these properties.

\begin{code}
module CompileDXT ( emitDXTransform
		  , emitControl
		  , genWorkProc
		  , newGUID
		  , newUUID

		  , mkObjectInfo
		  , ObjectInfo(..)
		  , genRegScript
		  , encodeProgID
		  
		  ) where

import Random (randomIO)
import Pretty
import PrettyUtils
import PanUtils (createNewGuid, shellOpen)
import Types
import IOExts
import List ( intersperse )
import Char ( isAlphaNum )
import WideString ( WideString(..), wideToStr )
import HDirect ( free )
import UI
import PanUtils ( unregisterProgID )
\end{code}

The DXTransform backend emits a bunch of different files - emitDXTransform,
takes care of it all:

\begin{code}
type Properties = [( String          -- property name
		   , String          -- label
		   , Type
		   , (String,String) -- min and max
		   , String          -- default value.
		   , UIElem
		   )]

emitDXTransform :: String
	        -> FilePath
		-> Properties
		-> [(Id,Type)]
		-> UIElem
		-> String
		-> Int
		-> IO ()
emitDXTransform name outDir props fields ui code noInputs = do
  oInfo' <- mkObjectInfo "DXImageTransform.Pan."
  let oInfo = oInfo'{className = mkClassName name}
  writeFile (outDir ++ "\\Transform.rgs")
  	    (genRegScript oInfo False)
  writeFile (outDir ++ "\\config.h")
  	    (genConfigInfo oInfo 1 noInputs props)
  writeFile (outDir ++ "\\PanTran2.h")
  	    (show (
	       genFieldDecls fields     $$
	       genPropMap oInfo props2  $$
	       initTransform props      $$
  	       vcat (map genSetPropStub props2) $$
  	       vcat (map genGetPropStub props2)) ++
	     '\n':code)
  writeFile (outDir ++ "\\Transform.idl")
	    (show $ genIDL oInfo props2 False)

  let testHTMLFile = outDir ++ '\\':{-"\\test" ++ -} name ++ ".html"
  randI <- randomIO
  let defImage = defaultImages !! (randI `mod` length defaultImages)
  writeFile testHTMLFile
  	    (show $ genHTML oInfo defImage ui)

  writeFile (outDir ++ "\\" ++ name ++ "_fx.xml")
	    (show (genXML oInfo props)) 
  writeFile (outDir ++ "\\" ++ name ++ "_ui.xml")
	    (show (genUI_XML oInfo props)) 
    -- to recycle some registry space, we unregister a server/component that is
    -- mapped to the same ProgID (but we won't delete it.)
    -- Hmm..this unregisters before we know whether compilation has succeeded or not;
    -- which is arguably a little bit too early.
  unregisterProgID (encodeProgID (progID oInfo (vProgID oInfo name) name))
  return ()
 where
  props2 = map (\ (x,_,y,_,_,_) -> (x,y)) props

  mkClassName = filter isAlphaNum


defaultImages =
  map (++ ".gif")
    ["piggy", "ducky", "bwcat", "dog"
    , "pineapple", "bananas", "erik", "pan-playing"
    , "flower", "flower-b", "escher", "oege"
     -- "jerzy"
     -- "pat", "charlotte", "becky", "jake", "kids-low"
    ]

\end{code}

\begin{code}
emitControl :: String
	    -> FilePath
	    -> Properties
	    -> [(Id,Type)]
	    -> UIElem
	    -> String
	    -> Int
	    -> IO ()
emitControl name outDir props fields ui code noInputs = do
  oInfo' <- mkObjectInfo "PanControl."
  let oInfo = oInfo'{className = mkClassName name}
  writeFile (outDir ++ "\\PanCtrl.rgs")
  	    (genRegScript oInfo True)
  writeFile (outDir ++ "\\config.h")
  	    (genConfigInfo oInfo 1 noInputs props)
  writeFile (outDir ++ "\\PanCtrl2.h")
  	    (show (
	       genFieldDecls fields     $$
	       genPropMap oInfo props2  $$
  	       vcat (map genSetPropStub props2) $$
  	       vcat (map genGetPropStub props2)) ++
	     '\n':code)
  writeFile (outDir ++ "\\PanControl.idl")
	    (show $ genIDL oInfo props2 True)
    -- to recycle some registry space, we unregister a server/component that is
    -- mapped to the same ProgID (but we won't delete it.)
    -- Hmm..this unregisters before we know whether compilation has succeeded or not;
    -- which is arguably a little bit too early.
  unregisterProgID (encodeProgID (progID oInfo (vProgID oInfo name) name))
  writeFile (outDir ++ "\\" ++ name ++ ".html")
	    (show $
	        text "<html><head><title>Control test page</title></head><body>" $$
  	        text "<object classid=\"clsid:" <> text (init (tail (clsid oInfo))) <> text "\">" $$
	        text "</object>" $$
	        text "</body></html>")
  return ()
 where
  props2 = map (\ (x,_,y,_,_,_) -> (x,y)) props

  mkClassName = filter isAlphaNum


\end{code}


Generate the registration script for a DXTransform component.

\begin{code}
genRegScript :: ObjectInfo -> Bool -> String
genRegScript oInfo forControl = show $
  vertScope (text "HKCR") $
    vertScope (pID <+> desc) 
        (text "CLSID" <+> bindString (text (clsid oInfo))) $$
    vertScope (vpID <+> desc) 
        (text "CLSID"  <+> bindString (text (clsid oInfo)) $$
	 text "CurVer" <+> bindString pID) $$
    vertScope (text "NoRemove CLSID")
        (vertScope (text ("ForceRemove " ++ clsid oInfo) <+> desc)
	    (text "ProgID" <+> bindString pID $$
	     text "VersionIndependentProgID" <+> bindString vpID $$
	     text "ForceRemove 'Programmable'" $$
	     vertScope (text "InprocServer32 = s '%MODULE%'")
	               (text "val ThreadingModel = s 'Both'") $$
	     (if forControl then
	         text "ForceRemove 'Control'" $$
	         --text "ForceRemove 'ToolboxBitmap32' s '%MODULE', 101'" $$
	          -- I suspect the single quotes aren't needed on the LHSs below,
	          -- but that's how the VS wizard generates it..
	      text "'MiscStatus' = s '0'" $$
	         vertScope empty (text "'1' = s '131473'") $$
	      text "'Version' = s '1.0'"
              else
	         empty) $$
             text "'TypeLib'" <+> bindString (text (libid oInfo))))
 where
  vpID   = text vpID_s
  vpID_s = encodeProgID $ vProgID oInfo (className oInfo)
  pID    = text (encodeProgID $ progID oInfo vpID_s (version oInfo))
  desc   = bindString (text (className oInfo))

  bindString d = text "= s '" <> d <> char '\''

encodeProgID :: String -> String
encodeProgID = map encodeChar 
 where
   -- ProgIDs can only contain alphanumeric chars + '_', so we have to
   -- map every non-alphanumeric character to '_' to not run afoul of that
   -- restriction.
  encodeChar x
    | isAlphaNum x || x == '.' = x
    | otherwise    = '_'


vertScope :: Doc -> Doc -> Doc
vertScope d1 d2 = d1 $$ char '{' $$ text "" $$ nest 4 d2 $$ char '}'
\end{code}

Some per-transform info needs to be communicated to the ATL wrapper that
isn't generated by Pan. Do this by way of a header file.

\begin{code}
genConfigInfo :: ObjectInfo               -- transform name (no spaces allowed.)
              -> Int                      -- number of image inputs required and expected.
	      -> Int                      -- major version number (minor is assumed to be 0.)
	      -> Properties               -- properties of object
              -> String
genConfigInfo oInfo version noInputs props = 
  show $ 
  vcat [ hashDef "InterfaceName"     (ifaceName oInfo (className oInfo))
       , hashDef "NO_IMAGE_INPUTS"   (show noInputs)
       , hashDef "LIBID_NAME"        ("LIBID_" ++ libName oInfo (className oInfo))
       , hashDef "CLSID_NAME"        ("CLSID_" ++ className oInfo)
       , hashDef "IID_NAME"          ("IID_" ++ ifaceName oInfo (className oInfo))
       , hashDef "EXTRA_MISC_FLAGS"  miscFlags
       , text ""
       ]
 where
   propNames = map (\ (x,_,_,_,_,_) -> x) props

   miscFlags 
     | "Quality" `elem` propNames = "DXTMF_QUALITY_SUPPORTED"
     | otherwise = "0"

hashDef x y = text "#define" <+> text x <+> text y
\end{code}

The @WorkProc()@ takes care of locking and unpacking the input & output surfaces
before invoking the generated display proc. After it returns, we copy the bits
to the output surface.

Generate the WorkProc by way of some rather ad-hoc CPP macros - makes it easier
to generate it all from within Haskell.

\begin{code}
genWorkProc :: [(String, String, String)] -> Doc
genWorkProc inputSurfaces = 
  text "DEFINE_WORKPROC()" $$
  nest 3 (vcat (inputs ++ 
                (text "INVOKE_TRANSFORM();"):closeInputs)) $$
  text "END_WORKPROC()"
 where
  inputs = zipWith mkSurface inputSurfaces [0..]
  closeInputs = map closeSurface inputSurfaces
  
  mkSurface (bits, width, height) no = 
    text "OPEN_READ_SURFACE" <> tuple [ text (show no)
                                      , asPointer bits
				      , asPointer width
				      , asPointer height 
				      , asUpdater bits
				      , asUpdater width
				      , asUpdater height
				      ] <> semi

  closeSurface (bits,_,_) =  text "CLOSE_READ_SURFACE" <> parens (asPointer bits) <> semi

asPointer x = text x -- ('&':x)
asUpdater x = text "update_" <> text x
tuple xss = parens (hcat (intersperse comma xss))
  
\end{code}

To encode names of the class, interface etc. to use when generating the IDL, introduce
the ObjectInfo type which enumerates the per-transform settable meta data.

\begin{code}
data ObjectInfo
 = ObjectInfo 
     { className    :: String  -- short name
     , classString  :: (String -> String)  
         -- Takes class name as argument, returns name
	 -- that'll show up in DXT browsers.
     , ifaceName    :: (String -> String)
         -- takes class name argument
     , ifaceString  :: (String -> String)
         -- takes interface name as argument
     , libName      :: (String -> String)  -- returns typelib name
         -- takes class name as argument.
     , libString    :: (String -> String)
     , iid          :: String -- in {...} form
     , iid_uuid     :: String -- in uuid(...) form
     , clsid        :: String
     , clsid_uuid   :: String
     , libid        :: String
     , libid_uuid   :: String
     , progID       :: (String -> String -> String)
     , vProgID      :: (String -> String)
     , version      :: String
     }

{-
  Create a new ObjectInfo record, filling it in with
  new UUIDs along with default values for the rest of
  the fields. 
  
  Using what 'mkObjectInfo' returns, you would then
  configure the name & string fields to fit, e.g.,
  
     oI <- mkObjectInfo
     return oI{className="MyTransform"}
-}
mkObjectInfo :: String -> IO ObjectInfo
mkObjectInfo progIDPrefix = do
  i_id    <- newGUID
  cls_id  <- newGUID
  lib_id  <- newGUID
  return $
    ObjectInfo
      { className    = "PanTransform"
      , classString  = \ cnm -> unwords [cnm, "Class"]
      , ifaceName    = \ cnm -> 'I':cnm
      , ifaceString  = \ inm -> unwords [inm, "Interface"]
      , libName      = \ cnm -> cnm ++ "Lib"
      , libString    = \ lnm -> unwords [lnm, "1.0 Type Library"]
      , iid          = i_id
      , iid_uuid     = wrap_uuid i_id
      , clsid        = cls_id
      , clsid_uuid   = wrap_uuid cls_id
      , libid        = lib_id
      , libid_uuid   = wrap_uuid lib_id
      , version      = ".1"
      , vProgID      = \ cnm -> progIDPrefix ++ cnm
      , progID       = \ vprogID vers -> vprogID ++ vers
      }
     
 where
   wrap_uuid ls@('{':xs) = "uuid(" ++ init xs ++ ")"

newUUID :: IO String
newUUID = do
  str <- newGUID
  case str of
    ('{':xs)  -> return ("uuid(" ++ init xs ++ ")")

newGUID :: IO String
newGUID = do
  wstr <- createNewGuid
  str  <- wideToStr wstr
  let (WideString pwstr) = wstr
  free pwstr
  return str
   
\end{code}


For parameterised DXTransforms, we map the parameterised values
onto properties of the object implementing the transform. 

\begin{code}
genIDL :: ObjectInfo
       -> [(String, Type)]
       -> Bool
       -> Doc
genIDL oInfo propList forControl = 
  text "// IDL auto-generated by the Pan image compiler" $$
  importIDL "oaidl.idl" $$
  importIDL "ocidl.idl" $$
  (if not forControl then importIDL "Dxtrans.idl" else text "#include \"olectl.h\"") $$
  typedef_enum $$
  iface_header $$
  char '{' $$
  nest 3 (vcat (map genProp propList)) $$
  char '}' $$ text ""  $$
  genDXTLibDecl oInfo
 where
  typedef_enum 
    | null propList = empty
    | otherwise     =
       text "typedef enum" <+> text (className oInfo) <> text "DISPID" <+> char '{' $$
            nest 2 (enum_tags (map fst propList)) $$
       char '}' <+> text (className oInfo) <> text "DISPID;" $$
       text ""

  enum_tags ls@(x:xs)
      | forControl = vsepWith comma (map (mkDispID oInfo) ls)
      | otherwise  = 
       vsepWith comma $
        ((mkDispID oInfo x <+> equals <+> text "DISPID_DXE_NEXT_ID") : 
	 map (mkDispID oInfo) xs)

  iface_header = 
    ppListVert
    	[ text "object"
	, text "dual"
	, text "pointer_default(unique)"
	, text (iid_uuid oInfo)
	, text "helpstring" <> 
	      parens (text (show (ifaceString oInfo (ifaceName oInfo (className oInfo)))))
	] $$
    text "interface" <+> text (ifaceName oInfo (className oInfo)) <+>
	 text (if forControl then ": IDispatch" else ": IDXEffect")

  genProp (propNm, propTy) 
   | knownProperty propNm = empty
   | otherwise = 
    text "[propget, id(" <> mkDispID oInfo propNm <> text ")]" <+>
      text "HRESULT" <+> text propNm <> 
      parens (text "[out,retval]" <> text (showCType propTy) <> char '*' <+> 
      text "pVal") <> semi $$
    text "[propput, id(" <> mkDispID oInfo propNm <> text ")]" <+>
      text "HRESULT" <+> text propNm <> 
      parens (text "[in]" <> text (showCType propTy) <+>  text "val") <> semi
    
mkDispID oInfo nm = 
  text "DISPID_" <> text (className oInfo) <> text ('_':nm)

  

genDXTLibDecl :: ObjectInfo -> Doc
genDXTLibDecl oInfo =
     ppListVert
       [ text (libid_uuid oInfo)
       , text "version(1.0)"
       , text "helpstring" <> 
       	   parens (text (show (libString oInfo (libName oInfo (className oInfo)))))
       ] $$
     text "library" <+> text (libName oInfo (className oInfo)) $$
     vertScope empty (
        importLib "stdole32.tlb" $$
        importLib "stdole2.tlb" $$
        ppListVert
	   [ text (clsid_uuid oInfo)
	   , text "helpstring" <>
       	       parens (text (show (classString oInfo (className oInfo))))
	   ] $$
	text "coclass" <+> text (className oInfo) $$
	vertScope empty (
	    text "[default]" <+> 
	    text "interface" <+> 
	    text (ifaceName oInfo (className oInfo)) <> semi) <> semi $$
	text "") <> semi

importLib :: String -> Doc
importLib x = text "importlib" <> parens (text (show x)) <> semi

importIDL :: String -> Doc
importIDL x = text "import" <+> text (show x) <> semi

\end{code}

Emit declarations for the property setter/getters.

\begin{code}
genFieldDecls :: [(Id,Type)] -> Doc
genFieldDecls fields = 
  text "private:" $$
  nest 3 (vcat (map mkDecl fields))
 where
  mkDecl (x,ty) = text (showCType ty) <+> text x <> semi

-- since we currently derive all Pan DXTransforms from IDXEffect,
-- we get a couple of properties for 'free'.
knownProperty :: String -> Bool
knownProperty nm = nm `elem` ["Progress", "Duration", "Quality"]

initTransform :: Properties -> Doc
initTransform ls = 
  text "" $$
  text "HRESULT" <+> text "OnSetup(DWORD dwFlags)" $$
  vertScope empty (vcat (map assignMember ls) $$ text "Init(); return S_OK;")
 where
  assignMember (nm, _, _, _, val, _)
     | not (knownProperty nm) = text "m_" <> text nm <+> equals <+> text val <> semi
     | otherwise = empty

genPropMap :: ObjectInfo -> [(String, Type)] -> Doc
genPropMap oInfo [] = 
  text "public:" $$
  text "BEGIN_PROP_MAP" <> parens (text cppClass) $$
  text "END_PROP_MAP()"
genPropMap oInfo props = 
  text "public:" $$
  text "BEGIN_PROP_MAP" <> parens (text cppClass) $$
  nest 3 (vcat $ map propEntry props) $$
  text "END_PROP_MAP()"
 where
   propEntry (nm, ty) 
     | knownProperty nm = empty
     | otherwise = 
         text "PROP_ENTRY" <> 
	 ppTuple
	    [ text (show nm)
	    , mkDispID oInfo nm
	    , text "CLSID_NULL" -- no support for property pages just yet.
	    ]


genSetPropStub :: (String, Type) -> Doc
genSetPropStub (nm, ty) 
 | knownProperty nm = empty
 | otherwise = 
  text "STDMETHODIMP" <+> text "put_" <> text nm <>
  parens (text (showCType ty) <+> text "val") $$
  vertScope empty (
    text "m_" <> text nm <+> equals <+> text "val" <> semi $$
    text "update_m_" <> text nm <> text "();" $$
    text "return" <+> text "S_OK" <> semi)

genGetPropStub :: (String, Type) -> Doc
genGetPropStub (nm, ty)
 | knownProperty nm = empty
 | otherwise = 
  text "STDMETHODIMP" <+> text "get_" <> text nm <>
  parens (text (showCType ty) <+> text "*pVal") $$
  vertScope empty (
    text "HRESULT hr" <+> equals <+> text "S_OK" <> semi $$
    text "if (!pVal) {" $$
    nest 2 (text "hr = E_POINTER;") $$
    text "} else {" $$
    nest 2 (text "*pVal = m_" <> text nm <> semi) $$
    char '}' $$
    text "return hr" <> semi)

-- name of the C++ class on the ATL wrapper side
cppClass :: String
cppClass = "CPanImageImpl"

\end{code}

Generating HTML to test the DXTransform (using IE.) This code is
modestly ugly at the moment; will tidy up later.

\begin{code}
genHTML :: ObjectInfo
        -> String
        -> UIElem
	-> Doc
genHTML oInfo defImage ui = vcat
	[ docHeader oInfo
	, theDecls
	, theTransform vals defImage oInfo
	, mkUI oInfo ui
	, predefinedFuns vals oInfo
	, docFooter
	]
 where 
  theDecls = jScriptBlock decls
  (decls, vals) = varDecls oInfo ui

docHeader oInfo = 
  text "<html>" $$
  text "<head>" $$
  text "<title>DXTransform tester</title>" $$
   -- ugly (ToDo: provide a little bit of CSS abstraction.)
  text "<style type=\"text/css\">" $$
  text "<!--" $$
  text "body {" $$
  nest 2 (text "font-family: Arial;" $$
  	  text "font-size:   12px;"  $$
	  text "background-color: #cdb3b1;") $$
  text "}" $$
  text "-->" $$
  text "</style>" $$
  text "</head>" $$
  text "<body>"  $$
  text "<h2>Effect name:" <+> text (className oInfo) <+> text "</h2>"
  
docFooter = text "</body></html>" $$ text ""

predefinedFuns ds oInfo = jScriptBlock $
   -- only accept numeric and '.' keystrokes.
  js_function (prefixClass oInfo (text "fnNumeric")) [] (
	text "if (event.keyCode == 45 || event.keyCode == 46 || ((event.keyCode-48 >= 0) && (event.keyCode-48 <= 9))) {" $$
	nest 2 (text "event.returnValue=true;") $$
	text "} else {" $$
	nest 2 (text "event.returnValue=false;") $$
	text "}") $$
  text "" $$
  js_function (prefixClass oInfo (text "fnUpdate")) [] (ppDecls (map mkApp ds))
 where
  mkApp (val, var) = 
  	js_assign (prefixClass oInfo (text "filterApp.filters.item(0)." <> text val))
	          (prefixClass oInfo (text var))

theTransform ds defImage oInfo = 
  let qdefImage = text ("\"" ++ defImage ++ "\"") in
  text "<table border=0 cellpadding=2>" $$
  text "<tr><img name=" <> prefixClass oInfo (text "filterApp src=") <> qdefImage $$
  text "style=\"filter:progid:" <> text ("DXImageTransform.Pan." ++ className oInfo) <> 
  {- ppTuple (map mkApp ds) <> -} text "\"></tr>" $$
  text "<tr><input type=\"text\" name=\"" <> prefixClass oInfo (text "imageSource\" value=") <> qdefImage <> text " size=28></tr>" $$
  text "<script for=" <> prefixClass oInfo (text "imageSource event=onchange>") $$
  prefixClass oInfo (text "filterApp.src=" <> prefixClass oInfo (text "imageSource.value;")) $$
  text "</script>" $$
  text "</table>" $$
  text ""
 where
   -- IE's 'auto-update' expression values doesn't seem to work with filters, so
   -- we have to manually do the update whenever any of the filter parameters changes.
  mkApp (val, var) = text val <> equals <> text "expression" <> parens (text var)

mkUI :: ObjectInfo -> UIElem -> Doc
mkUI oInfo (HBox lab ls) = vcat (map (mkUI oInfo) ls)
mkUI oInfo (VBox lab ls) = vcat (map (mkUI oInfo) ls)
mkUI oInfo (Resizeable _ _ u) = mkUI oInfo u
mkUI oInfo (Slider "Time" _ dVal vNm) = createTimerUI oInfo "Time" dVal
mkUI oInfo (Slider lab (min,max) dVal nm) = 
   createSliderUI oInfo lab False (0,100) (min,max) dVal (round ((dVal-min) * 100/(max-min))) nm
mkUI oInfo (ISlider lab (min,max) dVal nm) =
   createSliderUI oInfo lab True (min,max) (min,max) dVal dVal nm

mkUI oInfo (RadioBox lab vals dVal nm) =
   -- ToDo: fix width calculation.
   text "<fieldset style=\"width:300;\">" $$
   text "<legend>" <> text lab <> text "</legend>" $$
   vcat (zipWith mkRItem vals [0..]) $$
   text "</fieldset></form>" $$
   text ""
  where
   mkRItem v l 
     = text "<input onclick=\"" <> js_assign (prefixClass oInfo (text (toCId lab))) (text (show l)) <>
       prefixClass oInfo (text "fnUpdate();\"") <+> 
       text "type=radio name=" <> prefixClass oInfo (text (toCId lab)) <+> text "value=" <> text (show l) <> checkMark <> 
       text v <> text "</input>"
      where
       checkMark 
        | l == dVal = text " checked>"
	| otherwise = char '>'

mkUI oInfo (Tick lab dVal nm) = 
  let t_lab = prefixClass oInfo (text (toCId lab)) in
  text "<input onclick=\"if (this.checked) " <> js_assign t_lab (text "1") <+> 
  text "else" <+> js_assign t_lab (text "0") <+> prefixClass oInfo (text "fnUpdate();\" type=checkbox name=") <> t_lab <+>
  (if dVal /= 0 then text "checked>" else text ">") <> 
  text lab <> text "</input>"
mkUI _ _ = empty

createSliderUI oInfo lab noMapping (smin,smax) (min,max) dVal dValSlider nm =
   text "<form>" $$
   text "<fieldset style=\"width:300;\">" $$
   text "<legend>" <> text lab <> text "</legend>" $$
   let objValue = prefixClass oInfo (text nm <> text ".value") in
   table (tabRow 
     [ sliderObject oInfo nm smin smax 200 20 dValSlider
     , jScriptBlock (
         -- validator for input field; ensuring that values are clamped.
        js_function (prefixClass oInfo (text "update_") <> text nm)
	            [ prefixClass oInfo (text nm)
		    , prefixClass oInfo (text "slider_" <>  text nm)
		    ] (
   	  text "if (" <> objValue <+> char '<' <+> text (show min) <+> text ") {" $$
	     nest 2 (js_assign objValue (text (show min))) $$
	  text "} else if (" <> objValue <+> char '>' <+> text (show max) <+> text ") {" $$
	     nest 2 (js_assign objValue (text (show max))) $$
	  text "}" $$
          js_assign (prefixClass oInfo (text (toCId lab))) objValue $$
	  js_assign (prefixClass oInfo (text "slider_" <>  text nm <> text ".value"))
	            (text "100.0*" <> parens (prefixClass oInfo (text (toCId lab)) <+> char '-' <+> text (show min)) <+>
		     char '/' <+> parens (text (show max) <+> char '-' <+> text (show min))) $$
          prefixClass oInfo (text "fnUpdate();"))) $$
          text "<input onkeypress=\"" <> prefixClass oInfo (text "fnNumeric()\" type=text name=") <>
                prefixClass oInfo (text nm) <+> text "value=" <> text (show dVal) <+> text "onchange=\"" <> prefixClass oInfo (text "update_") <> text nm <> ppTuple [prefixClass oInfo (text nm), prefixClass oInfo (text "slider_" <>  text nm)] <> text ";\"" <+> text "size=10>"
     ]) $$
   text "</fieldset></form>" $$
   text "<script for=" <> prefixClass oInfo (text "slider_") <> text nm <+> text "event=Scroll>" $$
   (if noMapping then
       js_assign objValue (text "this.Value") $$
       js_assign (prefixClass oInfo (text (toCId lab))) (text "this.Value")
    else
       js_assign (text "val")
	     (text (show min) <+> char '+' <+> text "this.Value" <> char '*' <>
	      parens (text (show max) <> text "- " <> text (show min)) <> text "/100.0") $$
      js_assign objValue (text "val") $$
     js_assign (prefixClass oInfo (text $ toCId lab)) (text "val")) $$
   prefixClass oInfo (text "fnUpdate();") $$
   text "</script>" $$
   text ""

createTimerUI oInfo nm defVal =
    jScriptBlock (
	   js_varDecl oInfo tickVarName "0" $$
	   js_function (prefixClass oInfo (text ("upd_"++nm))) []
	       (js_assign (prefixClass oInfo (text nm)) (tickVarName_pp <+> text "*0.03") $$
		js_assign (tickVarName_pp) (tickVarName_pp <> text "+1") $$
	        prefixClass oInfo (text "fnUpdate();"))) $$
     text "<script for=window event=onload language=\"javascript\">" $$
     text "window.setInterval(\"" <> prefixClass oInfo (text "upd_") <> text nm <> text "()\", 30);" $$
     text "</script>"
 where
  tickVarName_pp = prefixClass oInfo (text tickVarName)
  tickVarName = nm ++ "_ticks"

prefixClass :: ObjectInfo -> Doc -> Doc
prefixClass oInfo d = text (className oInfo) <> char '_' <> d

varDecls :: ObjectInfo -> UIElem -> (Doc, [(String, String)])
varDecls oInfo (HBox lab ls) = let (ds, sss) = unzip (map (varDecls oInfo) ls) in (vcat ds, concat sss)
varDecls oInfo (VBox lab ls) = let (ds, sss) = unzip (map (varDecls oInfo) ls) in (vcat ds, concat sss)
varDecls oInfo (Resizeable _ _ u) = varDecls oInfo u
varDecls oInfo (Slider lab (min,max) dVal nm) = 
 (js_varDecl oInfo (toCId lab) (show dVal), [(toCId lab, toCId lab)])
varDecls oInfo (ISlider lab (min,max) dVal nm) = 
 (js_varDecl oInfo (toCId lab) (show dVal), [(toCId lab, toCId lab)])
varDecls oInfo (RadioBox lab _ dVal nm) = 
 (js_varDecl oInfo (toCId lab) (show dVal), [(toCId lab, toCId lab)])
varDecls oInfo (Tick lab dVal nm) = 
 (js_varDecl oInfo (toCId lab) (show dVal), [(toCId lab, toCId lab)])
varDecls _ _ = (empty, [])

-- some helpers to aid in the construction of valid JScript code:
js_function name params body = 
   text "function" <+> name <> ppTuple params <+> text "{" $$
   nest 2 body $$
   char '}'

js_assign var val = var <+> equals <+> val <> semi

jScriptBlock body =
 text "<script language=javascript>" $$
 body $$
 text "</script>" $$
 text ""

js_varDecl oInfo nm val = text "var" <+> prefixClass oInfo (text nm) <+> equals <+> text val <> semi

sliderObject :: ObjectInfo -> String -> Int -> Int -> Int -> Int -> Int -> Doc
sliderObject oInfo nm min max w h defVal =
   text "<object id=" <> prefixClass oInfo (text "slider_") <> text nm $$
   text "classid=\"clsid:F08DF954-8592-11D1-B16A-00C0F0283628\"" $$
   text "width="  <> text (show w) $$
   text "height=" <> text (show h) $$
   text ">" $$
   text "<param name=\"Min\" value=" <> text (show min) <> text ">" $$
   text "<param name=\"Max\" value=" <> text (show max) <> text ">" $$
   text "<param name=\"TickStyle\" value=\"3\">" $$
   text "<param name=\"Value\" value=\"" <> text (show defVal) <> text "\">" $$
   text "</object>"

tabRow :: [Doc] -> Doc
tabRow ls = 
  text "<tr>" $$
  (if null ls then empty else (text "<td>" <> vsepWith (text ("</td><td>")) ls <> text "</td>")) $$
  text "</tr>"

table :: Doc -> Doc
table ls = 
  text "<table>" $$
  ls $$
  text "</table>"

\end{code}

Descriptive

\begin{code}
genXML :: ObjectInfo
       -> Properties
       -> Doc
genXML oInfo props = 
  xmlTag "Filter" [ xmlAttr "name" (className oInfo)
                  , xmlAttr "reg"  ("clsid:" ++ trimCurlies (clsid oInfo))
                  , xmlAttr "dllname" (className oInfo ++ ".dll")
                  ] (
    text "<Icon src=\"x-theme://ndmedia.microsoft.com/images/" <> text (className oInfo) <> text ".jpg\"/>" $$
    vsep (map mkParam props)) $$
  text ""
 where
  mkParam (nm,lab,ty,(min,max),def,_) =
        xmlSTag "Param" 
		[ xmlAttr "name" nm
	        , xmlAttr "type" (showCType ty)
	        , xmlAttr "def"  def
	        , xmlAttr "min"  min
	        , xmlAttr "max"  max
		]

trimCurlies ('{':xs) = init xs

genUI_XML :: ObjectInfo
          -> Properties
          -> Doc
genUI_XML oInfo props = 
  xmlTag "Preset"
	 [ xmlAttr "label" (className oInfo)
         , xmlAttr "reg"   ("clsid:" ++ trimCurlies (clsid oInfo))
	 , xmlAttr "icon"  "x-theme://ndmedia.microsoft.com/images/GenericIcon.gif"
	 ] (vsep (map mkUIParam props)) $$
  text ""
 where
  mkUIParam (nm,lab,ty,(min,max),def, kind) =
    case kind of
       Slider{} -> xmlSTag "Slider"
                           [ xmlAttr "label" lab
			   , xmlAttr "name"  nm
			   , xmlAttr "def"   def
			   , xmlAttr "min"   min
			   , xmlAttr "max"   max
			   ]
       ISlider{} -> xmlSTag "Slider"
                           [ xmlAttr "label" lab
			   , xmlAttr "name"  nm
			   , xmlAttr "def"   def
			   , xmlAttr "min"   min
			   , xmlAttr "max"   max
			   , xmlAttr "int"   "0" -- The slider is int-based.
			   ]
       Tick{} -> xmlSTag "Checkbox" 
		           [ xmlAttr "label" lab
			   , xmlAttr "name"  nm
	                   , xmlAttr "def"   def
		           ]
       RadioBox{} -> -- no support for radioboxes yet, so map it to a slider.
                 xmlSTag "Slider"
                           [ xmlAttr "label" lab
			   , xmlAttr "name"  nm
			   , xmlAttr "def"   def
			   , xmlAttr "min"   min
			   , xmlAttr "max"   max
			   , xmlAttr "int"   "0" -- see above comment.
			   ]
       _ -> empty

xmlTag :: String -> [Doc] -> Doc -> Doc
xmlTag nm attrs d =
    char '<' <> text nm <+> hsep attrs <> char '>' $$ nest 2 d $$
    text "</" <> text nm <> char '>'
	  
xmlSTag :: String -> [Doc] -> Doc
xmlSTag nm attrs = char '<' <> text nm <+> hsep attrs <> text "/>"

xmlAttr nm val = text nm <> equals <> text (show val)
\end{code}
